
#include "analysisFunction.h"
#include "shared.h"

double **gen_probs(){
  double **ret =new double*[256];
  for(int i=0;i<256;i++)
    ret[i] = new double[16];

  

  for(int i=0;i<256;i++){
    double p1 = log(1-pow(10,-i/10.0));
    double p3 = log((1-exp(p1))/3.0);
    double val= std::max(p1,p3);
    val=std::min(p1-val,p3-val);
    for(int ii=0;ii<16;ii++){
      ret[i][ii] = val;
    }
    ret[i][0]=ret[i][5]=ret[i][10]=ret[i][15] =0;
  }
  
  return ret;
}


class hetplas:public general{
private:
  double **probs;
  FILE *fp;
  int minQ;
public:
  int doHetPlas;
  int doNew1(tNode*,double**,int*);
  void doNew(funkyPars *pars);
  hetplas(const char *outfiles,argStruct *arguments,int inputtype);
  ~hetplas();
  void getOptions(argStruct *arguments);
  void run(funkyPars  *pars);
  void clean(funkyPars *pars);
  void print(funkyPars *pars);
  void printArg(FILE *argFile);
  
  
};
void hetplas::printArg(FILE *fp){
  fprintf(fp,"--------------\n%s:\n",__FILE__);
  fprintf(fp,"\t-doHetPlas=%d\n",doHetPlas);
  fprintf(fp,"\t-minQ=%d\n",minQ);
}

int hetplas::doNew1(tNode *tn,double **liks,int *oldsize){
  if(*oldsize<tn->l){
    if(*oldsize>0){
      fprintf(stderr,"freeing: oldsize=%d newsize=%d\n",*oldsize,tn->l);
      for(int i=0;i<4;i++)
	delete [] liks[i];
      fprintf(stderr,"done freeing\n");
    }
    for(int i=0;i<4;i++)
      liks[i] = new double[tn->l];
    *oldsize = tn->l;
  }
  int depth =0;
  for(int l=0;l<tn->l;l++)
    if(refToInt[tn->seq[l]]!=4&&tn->qs[l]>=minQ){
      for(int ll=0;ll<4;ll++){
      	//fprintf(stderr,"i*4+ll=%d qs=%d ofs=%d reftoInt=%d \n",i*4+ll,tn->qs[l],refToInt[tn->seq[l]]*4+ll,refToInt[tn->seq[l]]*4);
	liks[ll][depth] = probs[tn->qs[l]][refToInt[tn->seq[l]]*4+ll];
      }
      depth++;
    }
  return depth;
}

double like(double *x,double **liks,int seqdepth){
  double totLik=0;
  for(int s=0;s<seqdepth;s++){
    double tmp=0;
    for(int i=0;i<4;i++)
      tmp+=x[i]*exp(liks[i][s]);
    totLik += log(tmp);
  }
  return totLik;
}

double em(double *x,double **liks,int seqdepth,int maxIter,double tole){
  int itr=0;
  double likev=-1e12;
  while(1){
    double newx[4]={0,0,0,0};
    memcpy(newx,x,sizeof(double)*4);
    for(int s=0;s<seqdepth;s++){
      double tsum=0;
      for(int i=0;i<4;i++)
	tsum += x[i]*exp(liks[i][s]);
      for(int i=0;i<4;i++)
	newx[i] += x[i]*exp(liks[i][s])/tsum;
    }
    for(int i=0;i<4;i++)
      newx[i] /= (1.0*seqdepth);
    double newlike=like(newx,liks,seqdepth);
    
    if(fabs(newlike-likev)<tole){
      likev=newlike;
      memcpy(x,newx,sizeof(double)*4);
      //fprintf(stderr,"breaking");
      break;
    }

    if(1&&(newlike<likev)){
      fprintf(stderr,"newlike=%f oldlike=%f\n",newlike,likev);
      exit(0);
    }
    assert(newlike>=likev);
    //fprintf(stderr,"lik=%f newlike=%f\n",likev,newlike);

    likev=newlike;
    memcpy(x,newx,sizeof(double)*4);
  }
  //  fprintf(stderr,"itr=%d like=%f par=(%f,%f,%f,%f)\n",itr,likev,x[0],x[1],x[2],x[3]);

}



void hetplas::doNew(funkyPars *pars){
  double **res= new double*[pars->numSites];
  
  double **liks = new double*[4];
  int oldsize =0;
  for(int s=0;s<pars->numSites;s++)
    if(pars->keepSites[s]){
      res[s] = new double[pars->nInd];  
      for(int i=0;i<pars->nInd;i++){
	int seqdepth=doNew1(&pars->chk->nd[s][i],liks,&oldsize);
#if 0
	fprintf(stderr,"seqdepth=%d\n",seqdepth);
	for(int b=0;b<4;i++){
	  for(int j=0;j<seqdepth;j++)
	    fprintf(fp,"%f\t",liks[b][j]);
	  fprintf(fp,"\n");
	}
	fclose(fp);
#endif
	double par[4]={0.25,0.25,0.23,0.3-0.05};
	//fprintf(stderr,"lik=%f\n",like(par,liks,seqdepth));
	em(par,liks,seqdepth,100,0.0000001);
	int maxBase=angsd::whichMax(par,4);
	assert(maxBase!=-1);//<- if vals are equal...
	double llhAlt = like(par,liks,seqdepth);
	par[0]=par[1]=par[2]=par[3]=0.0;
	par[maxBase]=1.0;
	double llhNeu=like(par,liks,seqdepth);
	res[s][i] = 2*llhNeu-2*llhAlt;
      }
      
    }
  for(int i=0;i<4;i++)
    delete [] liks[i];
  delete [] liks;
  pars->extras[index] = res;
}

void hetplas::run(funkyPars *pars){
 
  if(!doHetPlas||pars->numSites==0)
    return;
  assert(pars->numSites>0);
  assert(pars->chk!=NULL);
  if(doHetPlas==1){
    double **res = new double*[pars->numSites];
    for(int s=0;s<pars->numSites;s++)
      if(pars->keepSites[s]){
	res[s] = new double[4*pars->nInd];
	for(int i=0;i<4*pars->nInd;i++)
	  res[s][i]=0;
	for(int i=0;i<pars->nInd;i++){
	  tNode *tn = &pars->chk->nd[s][i];
	  for(int l=0;l<tn->l;l++)
	    for(int ll=0;ll<4;ll++){
	      if(refToInt[tn->seq[l]]!=4&&tn->qs[l]>=minQ)
		//fprintf(stderr,"i*4+ll=%d qs=%d ofs=%d reftoInt=%d \n",i*4+ll,tn->qs[l],refToInt[tn->seq[l]]*4+ll,refToInt[tn->seq[l]]*4);
		res[s][i*4+ll] += probs[tn->qs[l]][refToInt[tn->seq[l]]*4+ll];
	    }
	}
	
      }
    pars->extras[index] = res;
  }
  if(doHetPlas==2){
    doNew(pars);
  }
  
   
}

void hetplas::clean(funkyPars *pars){
  if(!doHetPlas)
    return;
  if(doHetPlas==1||doHetPlas==2){
    double **tmp =(double **) pars->extras[index];
    for(int i=0;i<pars->numSites;i++)
      if(pars->keepSites[i])
	delete [] tmp[i];
    
    delete [] tmp;
  }
}

void hetplas::print(funkyPars *pars){
  if(!doHetPlas)
    return;
  if(doHetPlas==1){
    double **tmp =(double **) pars->extras[index];
    for(int i=0;i<pars->numSites;i++)
      if(pars->keepSites[i]){
	fprintf(fp,"%s\t%d",header->name[pars->refId],pars->posi[i]);
	double *lik = tmp[i]; 
	for(int ii=0;ii<4*pars->nInd;ii++)
	  fprintf(fp,"\t%f",lik[ii]);
	fprintf(fp,"\n");
      }
  }
  if(doHetPlas==2){
    double **tmp =(double **) pars->extras[index];
    for(int i=0;i<pars->numSites;i++)
      if(pars->keepSites[i]){
	fprintf(fp,"%s\t%d",header->name[pars->refId],pars->posi[i]);
	for(int ii=0;ii<pars->nInd;ii++)
	  fprintf(fp,"\t%f",tmp[i][ii]);
	fprintf(fp,"\n");
      }

  }
}


void hetplas::getOptions(argStruct *arguments){
  //from command line
  

  doHetPlas=angsd::getArg("-doHetPlas",doHetPlas,arguments);
 
  if(doHetPlas==0)
    return;

  minQ = angsd::getArg("-minQ",minQ,arguments);
  
  printArg(arguments->argumentFile);

}


hetplas::hetplas(const char *outfiles,argStruct *arguments,int inputtype){
  doHetPlas =0;
  minQ = 13;
  probs=NULL;
 
  
  if(arguments->argc==2){
    if(!strcmp(arguments->argv[1],"-doHetPlas")){
      printArg(stdout);
      exit(0);
    }else
      return;
  }


  getOptions(arguments);
  if(doHetPlas)
    fprintf(stderr,"running doHetPlas=%d\n",doHetPlas);
  if(doHetPlas){
    probs=gen_probs();
    fp=openFile(outfiles,".hetGL");

  } 
}

hetplas::~hetplas(){
  if(doHetPlas)
    fclose(fp);

}
